/*=============================================================================
Vitalnix User Management Suite
accdb/accdb.c - ACCDB frontend
  Copyright © Jan Engelhardt <jengelh [at] linux01 gwdg de>, 2003 - 2005
  -- License restrictions apply (LGPL v2)

  This file is part of Vitalnix.
  Vitalnix is free software; you can redistribute it and/or modify it
  under the terms of the GNU Lesser General Public License as published
  by the Free Software Foundation; however ONLY version 2 of the License.

  Vitalnix is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
  General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this program kit; if not, write to:
  Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
  02111-1307, USA.

  -- For details see the LICENSE file
=============================================================================*/
#include <ctype.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <libHX.h>
#include <accdb.h>
#include "accdb/bf.h"

#ifndef _WIN32
#    ifndef __USE_GNU
#        define GNU_WAS_NOT_IN 1 // dirty hack
#        define __USE_GNU 1
#    endif
#    define __USE_GNU 1 // dirty hack
#    include <crypt.h>
#    ifdef GNU_WAS_NOT_IN
#        undef __USE_GNU
#    endif
#    undef GNU_WAS_NOT_IN
#endif

#define MAXFNLEN 256  // max length for filename buffer
#define MAXLNLEN 1024 // max length for line buffer

struct opt {
    char *def_backend;
    char *nmod, *smod;
};

static int genpw_random(char *, size_t, unsigned long, void *);
static int genpw_phonemic(char *, size_t, unsigned long, void *);
inline static int ld_cleanup(struct opt *, struct accdb_module *, int);
static void *ld_get_handle(const char *);
inline static void ld_get_submod(const char *, struct opt *);
static int ld_get_symbols(struct accdb_module *);
static void read_config(struct opt *, int);

//-----------------------------------------------------------------------------
struct accdb_module *accdb_load(const char *imod) {
    struct accdb_module *mp = NULL;
    struct opt cfg;
    int eax;

    memset(&cfg, 0, sizeof(struct opt));
    read_config(&cfg, 0);

    if((strcmp(imod, "*") == 0 && (imod = cfg.def_backend) == NULL)) {
        errno = ld_cleanup(&cfg, mp, errno);
    }

    if(*imod == '\0') {
        errno = ld_cleanup(&cfg, NULL, EINVAL);
        return NULL;
    }

    ld_get_submod(imod, &cfg);
    if((mp = malloc(sizeof(struct accdb_module))) == NULL ||
     (mp->handle = ld_get_handle(cfg.nmod)) == NULL ||
     ld_get_symbols(mp) <= 0) {
        errno = ld_cleanup(&cfg, mp, errno);
        return NULL;
    }

    if(mp->init != NULL && (eax = mp->init(mp, cfg.smod)) <= 0) {
        errno = ld_cleanup(&cfg, mp, eax);
        return NULL;
    }

    ld_cleanup(&cfg, NULL, 0);
    return mp;
}

void accdb_unload(struct accdb_module *mp) {
    if(mp->deinit != NULL) { mp->deinit(mp); }
    ld_cleanup(NULL, mp, 0);
    return;
}

int accdb_genpw(char *plain, size_t len, unsigned long flags) {
    void *randp = HX_rand_init(HXRAND_DEV_URANDOM | HXRAND_SYS_RAND);
    if(randp == NULL) { return 0; }

    if(flags & GENPW_PHONEMIC) {
        genpw_phonemic(plain, len, flags & ~GENPW_PHONEMIC, randp);
    } else {
        genpw_random(plain, len, flags, randp);
    }

    HX_rand_deinit(randp);
    return 1;
}

int accdb_cryptpw(const char *key, const char *salt, int meth,
 char **crypted)
{
    void *randp = HX_rand_init(HXRAND_DEV_URANDOM | HXRAND_SYS_RAND);
    static const char salt_pool[] =
     "./0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz";
    void *rx = NULL;
    int rv = 1;

    if(randp == NULL) { return -errno; }

    switch(meth) {
#ifndef _WIN32
        case CRYPW_DES: {
            struct crypt_data cd;
            memset(&cd, 0, sizeof(struct crypt_data));
            if(salt == NULL) {
                char my_salt[3], *saltp = my_salt;
                int n = 2;
                while(n--) { *saltp++ = salt_pool[HX_irand(randp, 0, 64)]; }
                *saltp++ = '\0';
                rx = crypt_r(key, my_salt, &cd);
            } else {
                rx = crypt_r(key, salt, &cd);
            }
            if(rx == NULL) { rv = 0; }
            else { *crypted = HX_strdup(rx); }
            break;
        }
        case CRYPW_MD5: {
            struct crypt_data cd;
            memset(&cd, 0, sizeof(struct crypt_data));
            if(salt == NULL) {
                char my_salt[13], *saltp = my_salt;
                int n = 8;

                memset(my_salt, 0, 13);
                *saltp++ = '$';
                *saltp++ = '1';
                *saltp++ = '$';
                while(n--) { *saltp++ = salt_pool[HX_irand(randp, 0, 64)]; }
                *saltp++ = '$';
                rx = crypt_r(key, my_salt, &cd);
            } else {
                rx = crypt_r(key, salt, &cd);
            }
            if(rx == NULL) { rv = 0; }
            else { *crypted = HX_strdup(rx); }
            break;
        }
#endif
        case CRYPW_BLOWFISH: {
            char res[64];
            if(salt == NULL) {
                char my_salt[64], *saltp = res;
                int n = 22;

                while(n--) { *saltp++ = salt_pool[HX_irand(randp, 0, 64)]; }
                *saltp++ = '\0';

                snprintf(my_salt, 64, "$2a$05$%s", res);
                rx = _crypt_blowfish_rn(key, my_salt, res, 64);
            } else {
                rx = _crypt_blowfish_rn(key, salt, res, 64);
            }
            if(rx == NULL) { rv = 0; }
            else { *crypted = HX_strdup(res); }
            break;
        }
        default:
            HX_rand_deinit(randp);
            return 0;
            break;
    }

    HX_rand_deinit(randp);
    return 1;
}

//-----------------------------------------------------------------------------
static int genpw_random(char *plain, size_t size, unsigned long fl, void *rd) {
    size_t cnt = size;
    char *ptrav = plain, *pcase = NULL;

    while(cnt--) {
        // 20% probability for a digit
        *ptrav++ = !HX_irand(rd, 0, 5) ? '0' + HX_irand(rd, 0, 10) :
         'a' + HX_irand(rd, 0, 26);
    }

    if(fl & GENPW_ONE_CASE) {
        pcase = &plain[HX_irand(rd, 0, size)];
        *pcase = toupper(*pcase);
    }

    if(fl & GENPW_ONE_DIGIT) {
        char *pdigit;
        // do not digitize the case letter
        while((pdigit = &plain[HX_irand(rd, 0, size)]) == pcase);
        *pdigit = '0' + HX_irand(rd, 0, 10);
    }

    *ptrav++ = '\0';
    return 1;
}

/*-----------------------------------------------------------------------------
Password generation function utilizing phonemic rules to generate secure,
but pronouncible and easy-memorizable passwords.

From ftp://ftp.gwdg.de/pub/suse/i386/8.2/suse/src/pwgen-2.02-68.src.rpm

Copyright (C) 2001, 2002 by Theodore Ts'o <tytso@alum.mit.edu>
This file [part] may be distributed under the terms of the GNU Public License.

The code was taken from pwgen-2.02/pw_phonemes.c and has been incorporated into
here and is also altered. The original author has nothing to do with this here,
so do not bug him if you find something not working.

Bugfixes I applied in here have been upstreamed in 2.03. */

static int genpw_phonemic(char *buf, size_t size, unsigned long ft_flags,
 void *randp)
{
    enum el_flags {
        FL_CONSONANT = 1 << 0,
        FL_VOWEL     = 1 << 1,
        FL_DIPTHONG  = 1 << 2,
        FL_NOT_FIRST = 1 << 3,
    };
    const struct pw_element {
        const char *str;
        const enum el_flags flags;
    } phonemic_element[] = {
        {"a",  FL_VOWEL},
        {"ae", FL_DIPTHONG | FL_VOWEL},
        {"ah", FL_DIPTHONG | FL_VOWEL},
        {"ai", FL_DIPTHONG | FL_VOWEL},
        {"b",  FL_CONSONANT},
        {"c",  FL_CONSONANT},
        {"ch", FL_CONSONANT | FL_DIPTHONG},
        {"d",  FL_CONSONANT},
        {"e",  FL_VOWEL},
        {"ee", FL_DIPTHONG | FL_VOWEL},
        {"ei", FL_DIPTHONG | FL_VOWEL},
        {"f",  FL_CONSONANT},
        {"g",  FL_CONSONANT},
        {"gh", FL_CONSONANT | FL_DIPTHONG | FL_NOT_FIRST},
        {"h",  FL_CONSONANT},
        {"i",  FL_VOWEL},
        {"ie", FL_DIPTHONG | FL_VOWEL},
        {"j",  FL_CONSONANT},
        {"k",  FL_CONSONANT},
        {"l",  FL_CONSONANT},
        {"m",  FL_CONSONANT},
        {"n",  FL_CONSONANT},
        {"ng", FL_CONSONANT | FL_DIPTHONG | FL_NOT_FIRST},
        {"o",  FL_VOWEL},
        {"oh", FL_DIPTHONG | FL_VOWEL},
        {"oo", FL_DIPTHONG | FL_VOWEL},
        {"p",  FL_CONSONANT},
        {"ph", FL_CONSONANT | FL_DIPTHONG},
        {"qu", FL_CONSONANT | FL_DIPTHONG},
        {"r",  FL_CONSONANT},
        {"s",  FL_CONSONANT},
        {"sh", FL_CONSONANT | FL_DIPTHONG},
        {"t",  FL_CONSONANT},
        {"th", FL_CONSONANT | FL_DIPTHONG},
        {"u",  FL_VOWEL},
        {"v",  FL_CONSONANT},
        {"w",  FL_CONSONANT},
        {"x",  FL_CONSONANT},
        {"y",  FL_CONSONANT},
        {"z",  FL_CONSONANT},
    };
#define NUM_ELEMENTS (sizeof(phonemic_element) / sizeof(struct pw_element))
#define rnd(__number) HX_irand(randp, 0, (__number))

    enum el_flags flags, prev, should_be;
    const char *str;
    size_t c, len;
    int first, i;

try_again:
    c         = 0;
    prev      = 0x0;
    should_be = 0x0;
    first     = 1;
    should_be = rnd(2) ? FL_VOWEL : FL_CONSONANT;

    while(c < size) {
        i     = rnd(NUM_ELEMENTS);
        str   = phonemic_element[i].str;
        len   = strlen(str);
        flags = phonemic_element[i].flags;

        // Filter on the basic type of the next element
        if((flags & should_be) == 0) { continue; }

        // Handle the FL_NOT_FIRST flag
        if(first && (flags & FL_NOT_FIRST)) { continue; }

        // Do not allow FL_VOWEL followed a FL_VOWEL/FL_DIPTHONG pair
        if((prev & FL_VOWEL) && (flags & (FL_VOWEL | FL_DIPTHONG)) ==
         (FL_VOWEL | FL_DIPTHONG)) {
            continue;
        }

        // Do not allow us to overflow the buffer
        if(len > size - c) { continue; }

        // OK, we found an element which matches our criteria, let 's do it!
        strcpy(buf + c, str);

        // Handle GENPW_ONE_CASE
        if((ft_flags & GENPW_ONE_CASE) && (first || (flags & FL_CONSONANT)) &&
         rnd(10) < 3) {
            buf[c] = toupper(buf[c]);
            ft_flags &= ~GENPW_ONE_CASE;
        }

        c += len;

        // Time to stop?
        if(c >= size) { break; }

        // Handle GENPW_ONE_DIGIT
        if((ft_flags & GENPW_ONE_DIGIT) && !first && rnd(10) < 3) {
            buf[c++] = rnd(10) + '0';
            buf[c] = '\0';
            ft_flags &= ~GENPW_ONE_DIGIT;
            first     = 1;
            prev      = 0x0;
            should_be = rnd(2) ? FL_VOWEL : FL_CONSONANT;
            continue;
        }

        // OK, figure out what the next element should be
        if(should_be == FL_CONSONANT) {
            should_be = FL_VOWEL;
        } else {
            // should_be == FL_VOWEL
            if((prev & FL_VOWEL) || (flags & FL_DIPTHONG) || rnd(10) > 3) {
                should_be = FL_CONSONANT;
            } else {
                should_be = FL_VOWEL;
            }
        }

        prev  = flags;
        first = 0;
    }

    if(ft_flags & (GENPW_ONE_CASE | GENPW_ONE_DIGIT)) {
        goto try_again;
    }
    return 1;
#undef NUM_ELEMENTS
#undef rnd
}

//-----------------------------------------------------------------------------
inline static int ld_cleanup(struct opt *c, struct accdb_module *mp, int ecx) {
    if(c != NULL) {
        read_config(c, 1);
        free(c->nmod);
    }
    if(mp != NULL) {
        if(mp->handle != NULL) { HX_dlclose(mp->handle); }
        free(mp);
    }
    return ecx;
}

static void *ld_get_handle(const char *module) {
    const char *ext[] = {".so", ".dll", "", NULL}, **extp = ext;
    void *h;

    // Try plain filename first (as does accdbinfo)
    if((h = HX_dlopen(module, HXLD_NOW)) != NULL) { return h; }

    // If not, prepend accdb_, append an extension and try again
    while(*extp != NULL) {
        char fn[MAXFNLEN];
        snprintf(fn, MAXFNLEN, "accdb_%s%s", module, *extp);
        if((h = HX_dlopen(fn, HXLD_NOW)) != NULL) { break; }
        ++extp;
    }

    return h;
}

inline static void ld_get_submod(const char *i, struct opt *c) {
    char *p;
    c->nmod = HX_strdup(i);
    c->smod = NULL;
    if((p = strchr(c->nmod, ':')) != NULL) {
        *p++ = '\0';
        c->smod = p;
    }
    return;
}

static int ld_get_symbols(struct accdb_module *mp) {
    mp->name      = HX_dlsym(mp->handle, "_module_name");
    mp->desc      = HX_dlsym(mp->handle, "_module_desc");
    mp->info      = HX_dlsym(mp->handle, "_module_info");
    mp->init      = HX_dlsym(mp->handle, "q_init");
    mp->open      = HX_dlsym(mp->handle, "q_open");
    mp->usertrav  = HX_dlsym(mp->handle, "q_usertrav");
    mp->userinfo  = HX_dlsym(mp->handle, "q_userinfo");
    mp->useradd   = HX_dlsym(mp->handle, "q_useradd");
    mp->usermod   = HX_dlsym(mp->handle, "q_usermod");
    mp->userdel   = HX_dlsym(mp->handle, "q_userdel");
    mp->grouptrav = HX_dlsym(mp->handle, "q_grouptrav");
    mp->groupinfo = HX_dlsym(mp->handle, "q_groupinfo");
    mp->groupadd  = HX_dlsym(mp->handle, "q_groupadd");
    mp->groupmod  = HX_dlsym(mp->handle, "q_groupmod");
    mp->groupdel  = HX_dlsym(mp->handle, "q_groupdel");
    mp->close     = HX_dlsym(mp->handle, "q_close");
    mp->sysctl    = HX_dlsym(mp->handle, "q_sysctl");
    mp->deinit    = HX_dlsym(mp->handle, "q_deinit");
    if(mp->open == NULL) { return -EFAULT; }
    return 1;
}

static void read_config(struct opt *config, int f) {
    const char *paths[] = {"accdb/../vetc", "/etc", NULL};
    struct rconfig_opt table[] = {
        {"DEFAULT_BACKEND", RCONF_STRING, &config->def_backend, NULL, NULL},
        {NULL},
    };
    if(f) { HX_rconfig_free(table); }
    else { HX_rconfig_pv(paths, "libaccdb/accdb", table, RCONFIG_ONE); }
    return;
}

//==[ End of file ]============================================================
