/*=============================================================================
Vitalnix User Management Suite
cui/main.c - Startup code
  Copyright © Jan Engelhardt <jengelh [at] linux01 gwdg de>, 2003 - 2005
  -- License restrictions apply (LGPL v2)

  This file is part of Vitalnix.
  Vitalnix is free software; you can redistribute it and/or modify it
  under the terms of the GNU Lesser General Public License as published
  by the Free Software Foundation; however ONLY version 2 of the License.

  Vitalnix is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
  General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this program kit; if not, write to:
  Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
  02111-1307, USA.

  -- For details see the LICENSE file
=============================================================================*/
#include <sys/file.h> // lockf()
#include <sys/stat.h>
#include <sys/types.h>
#include <errno.h>
#include <fcntl.h>
#include <locale.h>
#include <popt.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <libHX.h>
#include <accdb.h>
#include "cui/global.h"
#include "cui/xml_in.h"

enum {
    UG_ABRT   = 1 << 0,
    UG_DEFL   = 1 << 1,
    UG_PROMPT = 1 << 2,
};

static void print_banner(void);
static void init_var(int, const char **);
static void init_applock(void);

static void ui_mainmenu(void);
static size_t ui_generic(const char *, char *, size_t,
 unsigned long opts, ...);

static void ui_sync(void);
static int ui_sync_get_log(char *, size_t);
static int ui_sync_prompt_cont_add(void);
static int ui_sync_prompt_cont_del(void);

static void ui_pwlist(void);
static int ui_pwlist_sty(char *, size_t);

static void ui_single(void);

static int count_args(const char **);
static int parse_options(int *, const char ***);
static int load_config(char *);
static int pconfig_aflush(const char *, unsigned char, void *, void *);
static int pconfig_crymeth(const char *, unsigned char, void *, void *);

// Variables
static int _main_enter_ui = 1;

struct optmap_s Opt;
struct accdb_module *mm_output = NULL;

//-----------------------------------------------------------------------------
int main(int argc, const char **argv) {
    int eax;
    init_var(argc, argv);
    init_applock();
    print_banner();

    if(!(eax = parse_options(&argc, &argv))) { return !eax; }
    if(_main_enter_ui) { ui_mainmenu(); }
    unlink(Opt.f_lock);
    return EXIT_SUCCESS;
}

static void print_banner(void) {
    if(Opt.nterm) {
        printf(
         "\e[1;32m"">" "\e[0;32m"">" "\e[1;30m""> " // ]]]
         "\e[1;37m""V" "\e[0;37m""italnix" "\n\e[32m -- \e[0m" // ]]]]
        );
    } else {
        printf(">>> Vitalnix\n -- ");
    }

    printf(
      "C-spark %s (%s)\n\n"
      "This program comes with ABSOLUTELY NO WARRANTY; it is free software,\n"
      "you are welcome to redistribute it under certain conditions;\n"
      "for details see the file named \"LICENSE\", which should have\n"
      "come with this program.\n\n",
      Opt.progv, Opt.progd
    );
    return;
}

static void init_var(int argc, const char **argv) {
    const char *p;
    memset(&Opt, 0, sizeof(Opt));
    Opt.progv = "v2.00";
    Opt.progd = __DATE__ " " __TIME__;

    if((p = getenv("TERM")) != NULL && (strcmp(p, "linux") == 0 ||
     strcmp(p, "xterm") == 0 || strncmp(p, "screen", 6) == 0 ||
     strcmp(p, "msys") == 0 || strcmp(p, "rxvt") == 0)) {
        ++Opt.nterm;
    }

#ifdef __unix__
    if(!isatty(fileno(stdin)) && !isatty(fileno(stdout)) &&
     (p = getenv("DISPLAY")) != NULL && *p) {
        HX_fsystem(HX_FSYSTEM_ARGV | HX_FSYSTEM_EXEC, "xterm", argv,
         "-geometry", "80x25", "-bg", "#002040", "-fg", "gray", "-e", NULL);
        // exec() obviously failed when we get here
        perror("HX_fsystem()");
        exit(EXIT_FAILURE);
    }

    if(!isatty(fileno(stdout))) { Opt.nterm = 0; }
#endif

    HX_strclone(&Opt.f_home, "/home");
    HX_strclone(&Opt.f_lock, "/var/run/cspark.pid");
    HX_strclone(&Opt.f_shell, "/bin/bash");
    HX_strclone(&Opt.f_skeld, "/var/lib/empty");
    Opt.b_pwphon  = 1;
    Opt.c_aflush  = AFLUSH_OFF;
    Opt.c_crymeth = CRYPW_BLOWFISH;
    Opt.c_split   = 0;
    Opt.i_pswdlen = 10;

    umask(S_IRWXG | S_IRWXO); // 0077
    load_config("/etc/spark.cfg");
    load_config("cui/../vetc/spark.cfg");
    load_config("spark.cfg");

    setlocale(LC_COLLATE, "");
    return;
}

static void init_applock(void) {
#ifndef _WIN32
    char buf[8];
    int fd;

    if((fd = open(Opt.f_lock, O_RDWR | O_CREAT, 0666)) < 0) {
        perror("lockfile open()");
        return;
    }

    if(lockf(fd, F_TLOCK, 0) < 0) {
        perror("C-spark already running? lockf() returned");
        return;
    }

    ftruncate(fd, 0);
    snprintf(buf, sizeof(buf), "%d\n", getpid());
    write(fd, buf, strlen(buf));
    fcntl(fd, F_SETFD, FD_CLOEXEC);
    // never close that file or the lock will be gone
#endif
    return;
}

int Q_OPEN(void) {
    int eax;

    if((mm_output = accdb_load(Opt.module_path)) == NULL) {
        fprintf(stderr, "Could not load ACCDB back-end: %s\n",
         strerror(errno));
        return 0;
    }

    if((eax = mm_output->open(mm_output)) <= 0) {
        fprintf(stderr, "Could not open ACCDB back-end: %s"
         " (eax=%d, errno=%d)\n", strerror(errno), eax, errno);
        accdb_unload(mm_output);
        return 0;
    }

    return 1;
}

void Q_CLOSE(void) {
    if(mm_output != NULL) {
        mm_output->close(mm_output);
        accdb_unload(mm_output);
    }
    return;
}

//-----------------------------------------------------------------------------
static size_t ui_generic(const char *msg, char *buf, size_t count,
 unsigned long opts, ...) {
    char defl[MAXLNLEN];
    va_list argp;

    va_start(argp, opts);
    *defl = '\0';
    printf("\n" "%s", msg);

    if(opts & UG_ABRT) {
        if(opts & UG_DEFL) {
            printf("(Use a single stand-alone dot to abort.)\n");
        } else {
            printf("(Leave blank to abort.)\n");
        }
    }

    if(opts & UG_PROMPT) {
        printf("%s ", va_arg(argp, const char *));
    }

    if(opts & UG_DEFL) {
        defl[MAXLNLEN - 1] = '\0';
        printf("[%s] ", strncpy(defl, buf, MAXLNLEN - 1));
    }

    printf("> ");
    fflush(stdout);

    if(fgets(buf, count, stdin) == NULL) { return 0; }
    HX_chomp(buf);
    if(opts & UG_DEFL) {
        if(*buf == '\0') {
            buf[count - 1] = '\0';
            strncpy(buf, defl, count - 1);
        } else if(buf[0] == '.' && buf[1] == '\0') {
            *buf = '\0';
            return 0;
        }
    }

    return strlen(buf);
}

static void ui_mainmenu(void) {
    char buf[MAXSNLEN];
    int m_running = 1;

    while(m_running) {
        printf("\n"
          "MAIN MENU\n"
          "=========\n"
          "<2>  Synchronize user lists\n"
          "<4>  Print password list\n"
          "<5>  Add a single user\n"
          "<0>  Quit\n"
          "\n"
          "CHOICE > "
        );

        fflush(stdout);
        fgets(buf, MAXSNLEN, stdin);

        switch(*buf) {
            case '2':
                ui_sync();
                break;
            case '4':
                ui_pwlist();
                break;
            case '5':
                ui_single();
                break;
            case '9':
            case '0':
            case 'q':
                m_running = 0;
                break;
        }
    }

    return;
}

//-----------------------------------------------------------------------------
static void ui_sync(void) {
    struct HXbtree *al = NULL, *lnlist = NULL;
    struct accdb_group dgroup;
    struct HXdeque *dl = NULL;
    char buf[MAXFNLEN];
    FILE *logp = NULL;

    printf("\n\n"
      "IMPORT USERS\n"
      "============\n"
    );

    if(!Q_OPEN()) { return; }

    if(!ui_generic(
     "Enter the group to which compare users from the Data Source with.\n",
     buf, MAXFNLEN, UG_ABRT | UG_PROMPT, "Destination group")) {
        printf("No group specified, aborting.\n");
        goto out;
    }

    if(!sync_prep_dgrp(&dgroup, buf)) { goto out; }

    if(!ui_generic(
     "Please enter the pathname to the file serving as input. The file must\n"
     "be in XML and must have the tags described in doc/sp_xmlds.html.\n",
     buf, MAXFNLEN, UG_ABRT | UG_PROMPT, "Input file")) {
        printf("No filename specified, aborting.\n");
        goto out;
    }

    if(sync_read_file(buf, &al) <= 0) { goto out; }

    sync_compare(&dgroup, al, &dl, &lnlist);
    sync_fixup_addl(al, lnlist);

    if(al->itemcount == 0 && dl->itemcount == 0) {
        printf("Already synchronized\n");
        goto out;
    }

    sync_set_aflush();
    if((logp = sync_open_log(buf, ui_sync_get_log(buf, MAXFNLEN))) == NULL) {
        goto out;
    }

    if(al->itemcount > 0 && ui_sync_prompt_cont_add() &&
     !sync_add(&dgroup, al, logp)) {
        printf("Error during operation. You might need to redo the"
         " Synchronization process\n" "to complete all jobs.\n");
        goto out2;
    }

    if(dl->itemcount > 0 && ui_sync_prompt_cont_del() &&
     !sync_del(dl, logp)) {
        printf("Error during operation. You might need to redo the"
         " synchronization process\n" "to complete all jobs.\n");
        //goto out2;
    }

out2:
    /* We actually need to flush, since we have added home directores in /home,
    thus it would be nice to have at least some users have their account
    working. (Well, just keep a backup of your DB files, operations-so-far are
    recorded in the logfile and could thus be reversed easily.)
    Hm, Q_CLOSE also does an implicit flush if none has been done yet... */
    Q_SYSCTL(ACCDB_FLUSHDB);

out:
    if(logp != NULL) { fclose(logp); }
    sync_free(al, dl);
    Q_CLOSE();
    return;
}

static int ui_sync_get_log(char *file, size_t count) {
    // Propose a default logfile name
    strncpy(file, "synclog-", count - 1);
    now_in_ymdhms(file + strlen(file), count - strlen(file));
    HX_strlcat(file, ".log", count);

    int eax = ui_generic(
      "Enter a name for a log file containing a preliminary list of added\n"
      "users and their respective plaintext passwords.\n",
      file, count, UG_ABRT | UG_DEFL | UG_PROMPT, "Log file"
    );

    if(eax == 0) { printf("No filename given, aborting.\n"); }
    if(eax <= 0) { return 0; }

    if(file_exists(file)) {
        char buf[MAXSNLEN];
        printf("\n"
          "File already exists. Specify what to do: [O]verwrite, "
          "[A]ppend or [C]ancel? (Leave blank to abort.)\n"
          "[Cancel] > "
        );

        fflush(stdout);
        fgets(buf, MAXSNLEN, stdin);
        HX_strlower(buf);
        if(*buf == 'a') { return 'a'; }
        else if(*buf == 'o') { return 'w'; }
        else if(*buf == '\0') {
            printf("No input, aborting.\n");
            return 0;
        } else {
            printf("You have to specify one of 'a', 'c' or 'o'. Aborting.\n");
            return 0;
        }
    }

    return 'w';
}

static int ui_sync_prompt_cont_add(void) {
    char buf[MAXSNLEN];
    printf("\n"
      "Continue with adding users? (Yes/No)\n"
      "[Yes] > "
    );

    fflush(stdout);
    fgets(buf, MAXSNLEN, stdin);
    HX_chomp(buf);
    HX_strlower(buf);

    if(*buf != '\0' && *buf != 'y') {
        printf("Skipping addition (interactive user request)\n");
        return 0;
    }

    return 1;
}

static int ui_sync_prompt_cont_del(void) {
    char buf[MAXSNLEN];
    printf("\n"
      "Continue with deleting users? (Yes/No)\n"
      "[Yes] > "
    );

    fflush(stdout);
    fgets(buf, MAXSNLEN, stdin);
    HX_chomp(buf);
    HX_strlower(buf);

    if(*buf != '\0' && *buf != 'y') {
        printf("Skipping deletion (interactive user request)\n");
        return 0;
    }

    return 1;
}

//-----------------------------------------------------------------------------
static void ui_pwlist(void) {
    char infile[MAXFNLEN], outfile[MAXFNLEN], style[MAXSNLEN], tpl[MAXFNLEN];
    struct pwl_param tmp;

    printf("\n\n"
     "PRINT PASSWORD LISTS\n"
     "====================\n"
    );

    if(!ui_generic("Enter the name of the logfile that was created during"
     " user list\n" "synchronization.\n", infile, MAXFNLEN,
     UG_ABRT | UG_PROMPT, "Input log file")) {
        printf("No filename specified, aborting.\n");
        return;
    }

    if(!ui_generic("Enter the name of the file where the formatted"
     " (\"beautified\")\n" "output should go.\n", outfile, MAXFNLEN,
     UG_ABRT | UG_PROMPT, "Output file")) {
        printf("No filename specified, aborting.\n");
        return;
    }

    if(!ui_pwlist_sty(style, MAXFNLEN)) { return; }

    if(pwl_needs_tpl(style) && !ui_generic("This style needs a template file."
     " Please enter the path to it.\n", tpl, MAXFNLEN, UG_ABRT | UG_PROMPT,
     "Template file")) {
        printf("No filename specified, aborting.\n");
        return;
    }

    tmp.in = infile;
    tmp.out = outfile;
    tmp.style = style;
    tmp.tpl = tpl;
    pwl_proc(&tmp);
    return;
}

static int ui_pwlist_sty(char *buf, size_t count) {
    printf("\n"
      "Enter the canoncial name of a function responsible for output\n"
      "formatting. Enter \"?\" (without quotes) to see a list of names.\n"
      "(Leave blank to abort.)\n"
    );

    while(1) {
        printf("Style > ");
        fflush(stdout);
        fgets(buf, count, stdin);
        HX_chomp(buf);
        if(*buf == '?') {
            pwl_styles();
            continue;
        } else if(*buf == '\0') {
            return 0;
        }
        break;
    }

    return 1;
}

//-----------------------------------------------------------------------------
static void ui_single(void) {
    char nname[MAXSNLEN], vname[MAXSNLEN], xuid[MAXSNLEN],
     sysgrp[MAXSNLEN];
    struct {
        const char *msg, *prompt;
        char *dest;
        size_t len;
    } q_bday = {
        "If you wish to use a scrambled birthdate as XUID,"
        " enter the date now, otherwise leave blank. Accepted"
        " formats are DD.MM.YYYY, MM/DD/YYYY, YYYY-MM-DD.\n",
        "Birthdate", xuid, MAXSNLEN,
    }, q_xuid = {
        "Enter the XUID of the new user (may be empty)\n",
         "External UID", xuid, MAXSNLEN,
    }, questions[] = {
        {"Enter the last name of the user to be added.\n", "Last name", nname, MAXSNLEN},
        {"Enter the first name of the user to be added.\n", "First name", vname, MAXSNLEN},
        {"Enter the group name or GID to add the user to.\n", "System group", sysgrp, MAXSNLEN},
        //{"Enter the subgroup of the new user (may be empty)\n", "Subgroup", subgrp, MAXSNLEN},
        {NULL},
    }, *travp = questions;
    struct user_param tmp;

    printf("\n\n"
     "ADD A SINGLE USER\n"
     "=================\n"
    );

    travp = questions;
    while(travp->msg != NULL) {
        if((ui_generic(travp->msg, travp->dest, travp->len,
         UG_ABRT | UG_PROMPT, travp->prompt)) <= 0) {
            goto out;
        }
        ++travp;
    }

enter_bday:
    if(ui_generic(q_bday.msg, q_bday.dest, q_bday.len,
     UG_PROMPT, q_bday.prompt) > 0) {
        if(bday_to_xuid(q_bday.dest, q_bday.dest, q_bday.len) == NULL) {
            fprintf(stderr, "Illegal date format\n");
            goto enter_bday;
        }
    } else {
        if(ui_generic(q_xuid.msg, q_xuid.dest, q_xuid.len,
         UG_PROMPT | UG_ABRT, q_xuid.prompt) <= 0) {
            goto out;
        }
    }

    tmp.vname = vname;
    tmp.nname = nname;
    tmp.xuid  = xuid;
    tmp.gname = sysgrp;
    autorun_user(&tmp);
out:
    return;
}

//-----------------------------------------------------------------------------
static int count_args(const char **args) {
    int argk = 0;
    while(*args++ != NULL) { ++argk; }
    return argk;
}

static int load_config(char *fn) {
    static struct rconfig_opt options_table[] = {
        {"AUTOFLUSH",      RCONF_CB,     NULL,                pconfig_aflush},
        {"DEFAULT_BACKEND", RCONF_STRING, &Opt.module_path,   NULL},
        {"ENCODING",       RCONF_STRING, &Opt.encoding,       NULL},
        {"HOME",           RCONF_STRING, &Opt.f_home,         NULL},
        {"MASTER_PREADD",  RCONF_STRING, &Opt.master_preadd,  NULL},
        {"MASTER_PREDEL",  RCONF_STRING, &Opt.master_predel,  NULL},
        {"MASTER_POSTADD", RCONF_STRING, &Opt.master_postadd, NULL},
        {"MASTER_POSTDEL", RCONF_STRING, &Opt.master_postdel, NULL},
        {"PSWD_LEN",       RCONF_INT,    &Opt.i_pswdlen,      NULL},
        {"PSWD_METH",      RCONF_CB,     NULL,                pconfig_crymeth},
        {"PSWD_PHON",      RCONF_BOOL,   &Opt.b_pwphon,       NULL},
        {"SHELL",          RCONF_STRING, &Opt.f_shell,        NULL},
        {"SKEL",           RCONF_STRING, &Opt.f_skeld,        NULL},
        {"SPLIT_LVL",      RCONF_UCHAR,  &Opt.c_split,        NULL},
        {"USER_PREADD",    RCONF_STRING, &Opt.user_preadd,    NULL},
        {"USER_PREDEL",    RCONF_STRING, &Opt.user_predel,    NULL},
        {"USER_POSTADD",   RCONF_STRING, &Opt.user_postadd,   NULL},
        {"USER_POSTDEL",   RCONF_STRING, &Opt.user_postdel,   NULL},
        {NULL},
    };

    return HX_rconfig(fn, options_table);
}

static int parse_options(int *argc, const char ***argv) {
#define LONG_OPTION_2(__a, __b) (((__a) << 8) | (__b))
    static const char *_empty_argv[] = {NULL};
    struct pwl_param tmp_P;
    struct sync_param tmp_S;
    struct user_param tmp_U;
    char *tmp;
    struct poptOption options_table[] = {
        /* Being a versatile (command line) tool needs a lot of options.
        Oh well. */

        // Password list printing
        {NULL, 'P', POPT_ARG_NONE, NULL, 'P',
         "Print password list (using any preceding --P* parameters)", NULL},
        {"Ph", 0, POPT_ARG_NONE, NULL,           LONG_OPTION_2('P', 'h'),
         "Show available styles", NULL},
        {"Pi", 0, POPT_ARG_STRING, &tmp_P.in,    LONG_OPTION_2('P', 'i'),
         "Input file (for -P)", "FILE"},
        {"Po", 0, POPT_ARG_STRING, &tmp_P.out,   LONG_OPTION_2('P', 'o'),
         "Output file (for -P)", "FILE"},
        {"Ps", 0, POPT_ARG_STRING, &tmp_P.style, LONG_OPTION_2('P', 's'),
         "Style to use (for -P)", NULL},
        {"Pt", 0, POPT_ARG_STRING, &tmp_P.tpl,   LONG_OPTION_2('P', 't'),
         "Template for style (if applies)", "FILE"},

        // User list synchronization
        {NULL, 'S', POPT_ARG_NONE, NULL, 'S',
         "Synchronize with list (using any preceding --S* parameters)", NULL},
        {"Sg", 0, POPT_ARG_STRING, &tmp_S.gname, LONG_OPTION_2('S', 'g'),
         "System group to synchronize against (for -S)", "NAME"},
        {"Si", 0, POPT_ARG_STRING, &tmp_S.in,    LONG_OPTION_2('S', 'i'),
         "XML Data Source (for -S)", "FILE"},
        {"So", 0, POPT_ARG_STRING, &tmp_S.out,   LONG_OPTION_2('S', 'o'),
         "Output log file (for -S)", "FILE"},

        // Single user addition
        {NULL, 'U', POPT_ARG_NONE, NULL, 'U',
         "Add single user (using any preceding --U* parameters)", NULL},
        {"Ub", 0, POPT_ARG_STRING, &tmp,          LONG_OPTION_2('U', 'b'),
         "Use provided birthdate as XUID (for -U)", "BDAY"},
        {"Ug", 0, POPT_ARG_STRING, &tmp_U.gname,  LONG_OPTION_2('U', 'g'),
         "System group to put this user in (for -U)", "GROUP"},
        {"Un", 0, POPT_ARG_STRING, &tmp_U.nname,  LONG_OPTION_2('U', 'n'),
         "Last name of the user (for -U)", NULL},
        {"Uv", 0, POPT_ARG_STRING, &tmp_U.vname,  LONG_OPTION_2('U', 'v'),
         "First name of the user (for -U)", NULL},
        {"Ux", 0, POPT_ARG_STRING, &tmp_U.xuid,   LONG_OPTION_2('U', 'x'),
         "Unique identifier for user (for -U)", "ID"},

        // Other options
        {NULL, 'M', POPT_ARG_STRING, &Opt.module_path, 0,
         "Use a different module than \"*\" (the default)", NULL},
        {NULL, 'c', POPT_ARG_STRING, &tmp, 'c',
         "Load alternate config file", "file"},
        {NULL, 'h', POPT_ARG_NONE, NULL, 'h', "Show this help message", NULL},
        {"version", 'v', POPT_ARG_NONE, NULL, 'v',
         "Show version number and exit", NULL},
        POPT_AUTOHELP
        POPT_TABLEEND
    };

    poptContext ctx;
    const char **args;
    int c, argk;

    memset(&tmp_P, 0, sizeof(struct pwl_param));
    memset(&tmp_S, 0, sizeof(struct sync_param));
    memset(&tmp_U, 0, sizeof(struct user_param));
    ctx = poptGetContext(**argv, *argc, *argv, options_table, 0);

    while((c = poptGetNextOpt(ctx)) >= 0) {
        switch(c) {
            /* Specifying any of the autorun options and/or commands disables
            entering the UI afterwards. */
            case LONG_OPTION_2('P', 'i'):
            case LONG_OPTION_2('P', 'o'):
            case LONG_OPTION_2('P', 's'):
            case LONG_OPTION_2('P', 't'):
            case LONG_OPTION_2('S', 'g'):
            case LONG_OPTION_2('S', 'i'):
            case LONG_OPTION_2('S', 'o'):
            case LONG_OPTION_2('U', 'g'):
            case LONG_OPTION_2('U', 'n'):
            case LONG_OPTION_2('U', 'v'):
            case LONG_OPTION_2('U', 'x'):
            // --Ph and --Ub are handled down below
                _main_enter_ui = 0;
                break;
            case 'P':
                _main_enter_ui = 0;
                autorun_pwl(&tmp_P);
                free(tmp_P.in);
                free(tmp_P.out);
                free(tmp_P.style);
                free(tmp_P.tpl);
                memset(&tmp_P, 0, sizeof(struct pwl_param));
                break;
            case LONG_OPTION_2('P', 'h'):
                _main_enter_ui = 0;
                pwl_styles();
                break;
            case 'S':
                _main_enter_ui = 0;
                autorun_sync(&tmp_S);
                free(tmp_S.gname);
                free(tmp_S.in);
                free(tmp_S.out);
                memset(&tmp_S, 0, sizeof(struct sync_param));
                break;
            case 'U':
                _main_enter_ui = 0;
                autorun_user(&tmp_U);
                free(tmp_U.gname);
                free(tmp_U.vname);
                free(tmp_U.nname);
                free(tmp_U.xuid);
                memset(&tmp_U, 0, sizeof(struct user_param));
                break;
            case LONG_OPTION_2('U', 'b'):
                _main_enter_ui = 0;
                tmp_U.xuid = bday_to_xuid(tmp, NULL, 0);
                free(tmp);
                break;
            case 'c':
                if(!load_config(tmp)) {
                     fprintf(stderr, "Parsing of config file %s"
                      " unsuccessful\n", tmp);
                }
                free(tmp);
                break;
            case 'h':
                poptPrintHelp(ctx, stdout, 0);
                exit(EXIT_SUCCESS);
                break;
            case 'v':
                printf("C-spark %s (%s)\n", Opt.progv, Opt.progd);
                exit(EXIT_SUCCESS);
                break;
        }
    }

    if(c < -1) {
        fprintf(stderr, "%s: %s\n",
         poptBadOption(ctx, 0), poptStrerror(c));
        poptFreeContext(ctx);
        return 0;
    }

    if((args = poptGetArgs(ctx)) != NULL) {
        argk = count_args(args);
        poptDupArgv(argk, args, argc, argv);
    } else {
        *argc = 0;
        *argv = _empty_argv;
    }

    return 1;
#undef LONG_OPTION_2
}

static int pconfig_aflush(const char *key, unsigned char type, void *ptr,
 void *uptr) {
    unsigned char *v = &Opt.c_aflush;
    if(stricmp(ptr, "off") == 0)          { *v = AFLUSH_OFF;     }
    else if(stricmp(ptr, "on") == 0)      { *v = AFLUSH_ON;      }
    else if(stricmp(ptr, "postadd") == 0) { *v = AFLUSH_POSTADD; }
    else { *v = AFLUSH_DEFAULT; }
    return 1;
}

static int pconfig_crymeth(const char *key, unsigned char type, void *ptr,
 void *uptr) {
    unsigned char *v = &Opt.c_crymeth;
    if(stricmp(ptr, "md5") == 0) {
        *v = CRYPW_MD5;
#ifdef _WIN32
        fprintf(stderr, "Warning: No MD5 support under Win32.\n");
#endif
    } else if(stricmp(ptr, "des") == 0) {
        *v = CRYPW_DES;
#ifdef _WIN32
        fprintf(stderr, "Warning: No DES support under Win32.\n");
#endif
    } else {
        *v = CRYPW_BLOWFISH;
    }
    return 1;
}

//==[ End of file ]============================================================
