/*=============================================================================
Vitalnix User Management Suite
cui/pwl.c - Print password lists
  Copyright © Jan Engelhardt <jengelh [at] linux01 gwdg de>, 2003 - 2005
  -- License restrictions apply (LGPL v2)

  This file is part of Vitalnix.
  Vitalnix is free software; you can redistribute it and/or modify it
  under the terms of the GNU Lesser General Public License as published
  by the Free Software Foundation; however ONLY version 2 of the License.

  Vitalnix is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
  General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this program kit; if not, write to:
  Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
  02111-1307, USA.

  -- For details see the LICENSE file
=============================================================================*/
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <libHX.h>
#include "cui/global.h"

#define PROT_RDTMPL struct pwl_style *S, struct pwl_param *P
#define PROT_HEADER struct pwl_style *S, FILE *fp, const char *file
#define PROT_THDR   struct pwl_style *S, FILE *fp, const char *sg
#define PROT_ENTRY  struct pwl_style *S, FILE *fp, const char *sg, \
  const char *nname, const char *vname, const char *user, const char *pw
#define PROT_TFOOT  struct pwl_style *S, FILE *fp
#define PROT_FOOTER struct pwl_style *S, FILE *fp

#define DEFINE_CAT_HEADER \
    struct HX_repmap catalog[] = { \
        {'f', NULL, (char *)file}, \
        {'v', NULL, Opt.progv}, \
        {0}, \
    }
#define DEFINE_CAT_THDR \
    struct HX_repmap catalog[] = { \
        {'g', NULL, (char *)sg}, \
        {'v', NULL, Opt.progv}, \
        {0}, \
    }
#define DEFINE_CAT_ENTRY \
    struct HX_repmap catalog[] = { \
        {'g', NULL, (char *)sg}, \
        {'n', NULL, (char *)nname}, \
        {'p', NULL, (char *)pw}, \
        {'u', NULL, (char *)user}, \
        {'v', NULL, (char *)vname}, \
        {0}, \
    }

enum {
    TPAR_HEADER = 0,
    TPAR_THDR,
    TPAR_ENTRY,
    TPAR_TFOOT,
    TPAR_FOOTER,
    TPAR_LASTMARK,
};

struct pwl_style {
    const char *name, *desc;
    int (*rd_tmpl)(PROT_RDTMPL);
    void (*header)(PROT_HEADER);
    void (*thdr)(PROT_THDR);
    void (*entry)(PROT_ENTRY);
    void (*tfoot)(PROT_TFOOT);
    void (*footer)(PROT_FOOTER);
    char *tp, *tpp[TPAR_LASTMARK];
};

// Function prototypes
static int pwl_proc2(struct pwl_param *);
static int pwl_proc3(struct pwl_param *);

static struct HXbtree *build_tree(FILE *);
static struct pwl_style *style_by_name(const char *);

// Prototypes for styles
static void sg_html_header(PROT_HEADER);
static void sg_html_thdr(PROT_THDR);
static void sg_html_entry(PROT_ENTRY);
static void sg_html_tfoot(PROT_TFOOT);
static void sg_html_footer(PROT_TFOOT);

static int sg_rtf_rdtmpl(PROT_RDTMPL);
static void sg_rtf_header(PROT_HEADER);
static void sg_rtf_thdr(PROT_THDR);
static void sg_rtf_entry(PROT_ENTRY);
static void sg_rtf_tfoot(PROT_TFOOT);
static void sg_rtf_footer(PROT_TFOOT);

static void sg_txt_thdr(PROT_THDR);
static void sg_txt_entry(PROT_ENTRY);
static void sg_txt_tfoot(PROT_TFOOT);

static void a1_header(PROT_HEADER);
static void a1_entry(PROT_ENTRY);

static int sb_rdtmpl(PROT_RDTMPL);
static void sb_header(PROT_HEADER);
static void sb_entry(PROT_ENTRY);
static void sb_footer(PROT_FOOTER);

// Variables
static struct pwl_style styles_table[] = {
    /* "sg": Per-sgroup sorted overview.
    Each subgroup (whose members will be sorted by name) gets its own table.
    Useful for the annual grade shift. */
    {name: "sg_html", desc: "Per-sgroup tables (HTML)", rd_tmpl: NULL,
    header: sg_html_header, thdr: sg_html_thdr, entry: sg_html_entry,
    tfoot: sg_html_tfoot, footer: sg_html_footer},

    {name: "sg_rtf", desc: "Per-sgroup tables (RTF)", rd_tmpl: sg_rtf_rdtmpl,
    header: sg_rtf_header, thdr: sg_rtf_thdr, entry: sg_rtf_entry,
    tfoot: sg_rtf_tfoot, footer: sg_rtf_footer},

    {name: "sg_txt", desc: "Per-sgroup tables (TXT)", rd_tmpl: NULL,
    header: NULL, thdr: sg_txt_thdr, entry: sg_txt_entry, tfoot: sg_txt_tfoot,
    footer: NULL},

    /* "a1": All-in-1 sorted overview.
    Input data is sorted by sgroup first, then name and is presented in one
    table. Useful for small user changes or administration-only. (Thus, a
    text/plain version will suffice. */
    {name: "a1", desc: "Single sorted view (TXT)", rd_tmpl: NULL,
    header: a1_header, thdr: NULL, entry: a1_entry, tfoot: sg_txt_tfoot,
    footer: NULL},

    /* "sb": Serienbrief (letter series)
    Originally brought into Dautool/Daxtraq in 2002 by Cristoph Thiel */
    {name: "sb", desc: "Serienbrief (letter series) (LATEX)",
    rd_tmpl: sb_rdtmpl, header: sb_header, thdr: NULL, entry: sb_entry,
    tfoot: NULL, footer: sb_footer},

    {NULL},
};

//-----------------------------------------------------------------------------
int pwl_proc(struct pwl_param *param) {
    struct pwl_style *sp = styles_table;

    if((sp = style_by_name(param->style)) == NULL) {
        fprintf(stderr, "No such style \"%s\": %s\n",
         param->style, strerror(errno));
        return 0;
    }

    if((param->ifp = fopen(param->in, "r")) == NULL) {
        fprintf(stderr, "Could not open %s: %s\n", param->in, strerror(errno));
        return -errno;
    }

    param->styp = sp;
    printf("Processing \"%s\" to \"%s\" using \"%s\"\n",
     param->in, param->out, param->style);
    return pwl_proc2(param);
}

int pwl_needs_tpl(const char *style) {
    struct pwl_style *p = style_by_name(style);
    if(p != NULL) { return p->rd_tmpl != NULL; }
    return 0;
}

void pwl_styles(void) {
    struct pwl_style *p = styles_table;
    printf("Available styles: (asterisk denotes \"requires template"
     " file\" using --Pt option)\n");
    while(p->name != NULL) {
        printf("\t" "%c%-10s   %s\n", (p->rd_tmpl == NULL) ? ' ' : '*',
         p->name, p->desc);
        ++p;
    }
    return;
}

//-----------------------------------------------------------------------------
static int pwl_proc2(struct pwl_param *param) {
    struct pwl_style *style = param->styp;
    int rv = 1;

    if(style->rd_tmpl != NULL && param->tpl == NULL) {
        fprintf(stderr, "--Pt argument needed for this style\n");
        return 0;
    }

    if((param->bt = build_tree(param->ifp)) == NULL) {
        fprintf(stderr, "build_tree(): %s\n", strerror(errno));
        fclose(param->ifp);
        return -errno;
    }

    fclose(param->ifp);
    if((param->ofp = fopen(param->out, "w")) == NULL) {
        fprintf(stderr, "Could not open %s: %s\n",
         param->out, strerror(errno));
        return -errno;
    }

    rv = pwl_proc3(param);
    fclose(param->ifp);
    HXbtree_free(param->bt);
    printf("\n");
    return rv;
}

static int pwl_proc3(struct pwl_param *param) {
    /* In 99% of all cases, administrators want (any kind of) output to be
    sorted, even if it is just some small number of users. Thus, the
    implemented "architecture" for realization includes callbacks for
    pwl_proc3(). */
    struct pwl_style *style = param->styp;
    struct HXbtree_node *nd;
    struct HXbtrav *travp;

    if(style->rd_tmpl != NULL) {
        int eax = style->rd_tmpl(style, param);
        if(eax < 0) { return eax; }
    }

    if(style->header != NULL) { style->header(style, param->ofp, param->out); }
    if((travp = HXbtrav_init(param->bt, NULL)) == NULL) {
        perror("pwl_proc3(): Initializing traverser failed");
        return -errno;
    }

    while((nd = HXbtraverse(travp)) != NULL) {
        struct HXbtrav *sg = HXbtrav_init(nd->data, NULL);
        struct HXbtree_node *sd;
        if(style->thdr != NULL) { style->thdr(style, param->ofp, nd->key); }

        while((sd = HXbtraverse(sg)) != NULL) {
            char *ptr = sd->key, *c_nname, *c_vname, *c_user, *c_pw;
            c_nname = HX_strsep(&ptr, ":");
            c_vname = HX_strsep(&ptr, ":");
            c_user  = HX_strsep(&ptr, ":");
            c_pw    = HX_strsep(&ptr, ":");
            if(style->entry != NULL) {
                style->entry(style, param->ofp, nd->key, c_nname, c_vname,
                 c_user, c_pw);
            }
        }

        if(style->tfoot != NULL) { style->tfoot(style, param->ofp); }
        HXbtree_free(nd->data);
    }

    if(style->footer != NULL) { style->footer(style, param->ofp); }
    HXbtrav_free(travp);
    return 1;
}

static struct HXbtree *build_tree(FILE *ifp) {
    /* This function reads a logfile as generated by the synchronization stage,
    puts it in a B-tree (again) just to retrieve it sorted afterwards. */
    struct HXbtree *bt;
    char buf[MAXLNLEN];
    int std = 0;

    bt = HXbtree_init(HXBT_MAP | HXBT_CKEY | HXBT_CMPFN, compare_wbc);
    if(bt == NULL) { return NULL; }

    while(fgets(buf, MAXLNLEN, ifp) != NULL) {
        char *bufp = buf, *c_type, *c_sgroup;
        struct HXbtree *sgp;

        HX_chomp(buf);
        if(*buf == '\0') { continue; }
        if(*buf == '#') {
            if(!std && strncmp(buf, "# $logformat", 12) == 0 &&
             strcmp(buf + 12, " " LOGFMT_VERSION) != 0) {
                fprintf(stderr, "Input file is not of version \""
                 LOGFMT_VERSION "\"\n");
                HXbtree_free(bt);
                return 0;
            }
            continue;
        }

        // Won't take any $logformat into account now since data has started.
        ++std;

        c_type = HX_strsep(&bufp, ":");
        c_sgroup = HX_strsep(&bufp, ":");
        if(*c_type != '+' || c_sgroup == NULL || bufp == NULL) { continue; }

        // Only print users which were added ('+')
        if((sgp = HXbtree_get(bt, c_sgroup)) == NULL) {
            sgp = HXbtree_init(HXBT_CDATA | HXBT_CMPFN, strcoll);
            HXbtree_add(bt, c_sgroup, sgp);
        }
        HXbtree_add(sgp, bufp);
    }

    return bt;
}

static struct pwl_style *style_by_name(const char *name) {
    struct pwl_style *p = styles_table;
    if(name == NULL) {
        errno = EINVAL;
        return NULL;
    }

    errno = ENOENT;
    while(p->name != NULL) {
        if(strcmp(name, p->name) == 0) { break; }
        ++p;
    }
    return p;
}

//-----------------------------------------------------------------------------
/* Please do not remove the "Formatted by" messages as they are the only
external advertisements for this software. */

static void sg_html_header(PROT_HEADER) {
    fprintf(fp, "<html><head><title>%s</title><style type=\"text/css\">"
     "body { font: 12pt \"Arial\", \"Sans Serif\"; }\n"
     "th { text-align: left; }\n"
     "</style><body>", file);
    return;
}

static void sg_html_thdr(PROT_THDR) {
    fprintf(fp, "<table width=\"100%%\"><tr><td><h1>Group %s</h1></td>"
     "<td align=\"right\"><i>Formatted by Vitalnix C-Spark %s</i></td></tr>"
     "</table><table cellpadding=\"5\" width=\"100%%\"><tr>"
     "<th width=\"50%%\">Name</th><th>Login</th><th>Password</th></tr>",
     sg, Opt.progv);
    return;
}

static void sg_html_entry(PROT_ENTRY) {
    char buf[MAXSNLEN];
    struct HX_repmap catalog[] = {
        {'n', NULL, buf},
        {'p', NULL, (void *)pw},
        {'u', NULL, (void *)user},
        {0},
    };
    snprintf(buf, MAXSNLEN, "%s, %s", nname, vname);
    HX_fstrrep(fp, "<tr><td>%n</td><td>%u</td><td>%p</td></tr>", catalog);
    return;
}

static void sg_html_tfoot(PROT_TFOOT) {
    fprintf(fp, "</table>");
    return;
}

static void sg_html_footer(PROT_FOOTER) {
    fprintf(fp, "</body></html>");
    return;
}

//-----------------------------------------------------------------------------
static int sg_rtf_rdtmpl(PROT_RDTMPL) {
    char *p;
    if((S->tp = read_file(P->tpl)) == NULL) {
        fprintf(stderr, "sg_rtf_rdtmpl(): Could not open template file"
         " %s: %s\n", P->tpl, strerror(errno));
        return 0;
    }

    /* The following fixup is so that the RTF file can be edited more easily.
    And of course, we keep it in one file, which is essential. */
    S->tpp[TPAR_HEADER] = p = S->tp;

    /* Micro$oft Office successfully ignores our private tags, so that we can
    check out what our template looks like without having to wade through. */

#define SEP(__str, __len, __ent) /* some small strstr_strsep() */ \
     p = strstr(p, (__str)); *p = '\0'; p += (__len); S->tpp[__ent] = p;
    SEP("{\\THDR}",   7, TPAR_THDR);
    SEP("{\\ENTRY}",  8, TPAR_ENTRY);
    SEP("{\\TFOOT}",  8, TPAR_TFOOT);
    SEP("{\\FOOTER}", 9, TPAR_FOOTER);
#undef SEP
    return 1;
}

static void sg_rtf_header(PROT_HEADER) {
    DEFINE_CAT_HEADER;
    HX_fstrrep(fp, S->tpp[TPAR_HEADER], catalog);
    return;
}

static void sg_rtf_thdr(PROT_THDR) {
    DEFINE_CAT_THDR;
    HX_fstrrep(fp, S->tpp[TPAR_THDR], catalog);
    return;
}

static void sg_rtf_entry(PROT_ENTRY) {
    DEFINE_CAT_ENTRY;
    HX_fstrrep(fp, S->tpp[TPAR_ENTRY], catalog);
    return;
}

static void sg_rtf_tfoot(PROT_TFOOT) {
    fprintf(fp, S->tpp[TPAR_TFOOT]);
    return;
}

static void sg_rtf_footer(PROT_FOOTER) {
    fprintf(fp, S->tpp[TPAR_FOOTER]);
    return;
}

//-----------------------------------------------------------------------------
static void sg_txt_thdr(PROT_THDR) {
    fprintf(fp, "\n\n"
     "                                  Formatted by Vitalnix C-Spark %s\n\n\n"
     "  >>> Group %s\n\n"
     "  %-35s  %-10s  %s\n"
     "  ====================================="
     "==================================\n",
     Opt.progv, sg, "Name", "Login", "Password");
    return;
}

static void sg_txt_entry(PROT_ENTRY) {
    char buf[MAXSNLEN];
    snprintf(buf, MAXSNLEN, "%s, %s", nname, vname);
    fprintf(fp, "  %-35s  %-10s  %s\n", buf, user, pw);
    return;
}

static void sg_txt_tfoot(PROT_TFOOT) {
    fputc(0x0C, fp);
    return;
}

//-----------------------------------------------------------------------------
static void a1_header(PROT_HEADER) {
    fprintf(fp, "\n\n"
     "                                  Formatted by Vitalnix C-Spark %s\n\n\n"
     "  %-8s  %-35s  %-10s  %s\n"
     "  ====================================="
     "==================================\n",
     Opt.progv, "Group", "Name", "Login", "Password");
    return;
}

static void a1_entry(PROT_ENTRY) {
    char buf[MAXSNLEN];
    snprintf(buf, MAXSNLEN, "%s, %s", nname, vname);
    fprintf(fp, "  %-8s  %-35s  %-10s  %s\n", sg, buf, user, pw);
    return;
}

//-----------------------------------------------------------------------------
static int sb_rdtmpl(PROT_RDTMPL) {
    /* Unlike sg_rtf, this template should only contain the
    THDR/ENTRY parts! (TFOOT is handled in sb_entry().) */
    if((S->tp = read_file(P->tpl)) == NULL) {
        fprintf(stderr, "sb_rdtmpl(): Could not open template file %s: %s\n",
         P->tpl, strerror(errno));
        return 0;
    }
    return 1;
}

static void sb_header(PROT_HEADER) {
    fprintf(fp,
      "\\documentclass[12pt]{article}\n\n"
      "\\usepackage{isolatin1}\n"
      "\\usepackage{graphics}\n"
      "\\usepackage{german}\n\n"
      "\\begin{document}\n\n"
    );
    return;
}

static void sb_entry(PROT_ENTRY) {
    DEFINE_CAT_ENTRY;
    HX_fstrrep(fp, S->tp, catalog);
    fprintf(fp, "\\newpage\n");
    return;
}

static void sb_footer(PROT_FOOTER) {
    fprintf(fp, "\\end{document}\n");
    free(S->tp);
    return;
}

//==[ End of file ]============================================================
