/*=============================================================================
Vitalnix User Management Suite
cui/sync.c - Where the real work happens
  Copyright © Jan Engelhardt <jengelh [at] linux01 gwdg de>, 2003 - 2005
  -- License restrictions apply (LGPL v2)

  This file is part of Vitalnix.
  Vitalnix is free software; you can redistribute it and/or modify it
  under the terms of the GNU Lesser General Public License as published
  by the Free Software Foundation; however ONLY version 2 of the License.

  Vitalnix is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
  General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this program kit; if not, write to:
  Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
  02111-1307, USA.

  -- For details see the LICENSE file
=============================================================================*/
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <libHX.h>
#include <accdb.h>
#include "cui/global.h"
#include "cui/xml_in.h"

inline static void free_addl_entry(struct ds_entry *);
static void handle_user(struct accdb_user *, struct HXbtree *,
 struct HXdeque *, long *);

//-----------------------------------------------------------------------------
int sync_prep_dgrp(struct accdb_group *grp, const char *name) {
    /* Here, we request a group to which compare the data source against.
    Usually, a group is always the same for a particular input set otherwise
    strange things may happen. You can either enter the group name or a GID.
    (Once there is a non-digit in your input, it is treated as a group
    name.) */
    int eax;

    if(only_digits(name)) {
        grp->gid = strtol(name, NULL, 0);
        grp->gname = NULL;
    } else {
        grp->gid = -1;
        grp->gname = (char *)name;
    }

    if((eax = Q_GROUPINFO(grp, grp, 1)) < 0) {
        fprintf(stderr, "Error querying the ACCDB: %s (eax=%d, errno=%d)\n",
         strerror(errno), eax, errno);
        return 0;
    } else if(eax == 0) {
        fprintf(stderr, "Group \"%s\" does not exist\n", name);
        return 0;
    }

    /* After [successful] groupinfo(), dgroup->gname is now a pointer to space
    malloc'ed from within the backend. So, dgroup->gname can safely be used
    after this line. */
    return 1;
}

int sync_read_file(const char *file, struct HXbtree **al) {
    /* In this part, we stuff all users from the data source into an
    associative array, called "al". Some elements of that array will further be
    changed until it is ready for running through import_ds(). */
    struct ds_entry *bucket;
    char buf[MAXSNLEN];
    void *state;
    int eax;

    if((eax = xmlds_open(file, &state)) <= 0) { return eax; }

    /* This loop fills the "al" associative array by generating a suited
    (pre-stage) login name. (These login names may already be used by the
    system and must first be fixed up (e.g. by adding a unique postcount
    number), so that there are only unique login name in the user database.) */

    *al = HXbtree_init(HXBT_MAP | HXBT_CKEY | HXBT_CMPFN, stricmp);

    while(1) {
        bucket = malloc(sizeof(struct ds_entry));

        if((eax = xmlds_read(state, bucket)) < 0) {
            fprintf(stderr, "Warning: Stumbled upon a bogus entry in"
             " Data Source, continuing\n");
            free(bucket);
            continue;
        } else if(eax == 0) {
            // End of list
            free(bucket);
            break;
        }

        complete_pentry(bucket, buf, MAXSNLEN);
        HXbtree_add(*al, buf, bucket);
    }

    xmlds_close(state);
    return 1;
}

void sync_compare(struct accdb_group *dgroup, struct HXbtree *al,
 struct HXdeque **dl, struct HXbtree **lnlist) 
{
    /* And this is now the comparison algorithm, (half of) the core. While
    traversing the users, we mark used login names by putting them in an array
    (lnlist). Login names of users which are about to be deleted are also
    pushed there, so they do not get reassigned right now, but at the next
    run-through. */
    struct accdb_user passwd;

    printf("\n" "Comparing Data Source against UDB\n");
    *lnlist = HXbtree_init(HXBT_CDATA | HXBT_SCMP);

    /* Sure, I could pack the if(dl != NULL) into the while() loop, but I do
    not do so, for performance reasons. Using the following approach, we only
    need to check for dl != NULL once, rather than for every user. */
    if(dl != NULL) {
        long ugrp = 0, ukeep = 0; // some stats

        *dl = HXdeque_init(0);
        while(Q_USERTRAV(&passwd)) {
            HXbtree_add(*lnlist, passwd.lname);

            if((dgroup->gname != NULL && passwd.igrp != NULL &&
             strcmp(dgroup->gname, passwd.igrp) == 0) ||
             (dgroup->gid != -1 && dgroup->gid == passwd.gid)) {
                // Users we look for (only specified group)
                ++ugrp;
                handle_user(&passwd, al, *dl, &ukeep);
            }
        }

        printf("%ld group member(s) found\n", ugrp);
        printf("    %ld to keep\n", ukeep);
        printf("    %lu to delete\n", (**dl).itemcount);
        printf("%lu to add\n", al->itemcount);
    } else {
        /* If "dl" is NULL, a single user is to be added, in which case nobody
        gets removed (moved from "al" to "dl") at all. We just fill up "lnlist"
        for login name fixup. */
        while(Q_USERTRAV(&passwd)) { HXbtree_add(*lnlist, passwd.lname); }
    }

    return;
}

void sync_fixup_addl(struct HXbtree *al, struct HXbtree *lnlist) {
    // Here comes the fixup loop (other part of the core).
    // DEVELOPER Notes: frees "lnlist"
    void *travp = HXbtrav_init(al, NULL);
    long stat_tries = 0, stat_users = 0;
    struct HXbtree_node *node;
    char sbuf[16];

    if(al->itemcount == 0) { return; }

    printf("Username fixup...\n");
    while((node = HXbtraverse(travp)) != NULL) {
        struct ds_entry *entry = node->data;
        int i = 0;

        sbuf[15] = '\0';
        strncpy(sbuf, entry->lname, 15);

        /* If the fixup'ed login name with the unique number already exists,
        bump the unique number. */
        while(HXbtree_find(lnlist, sbuf) != NULL) {
            fixup_login_name(entry->lname, ++i, sbuf, 16);
            ++stat_tries;
        }

        if(i > 0) { ++stat_users; }
        if(strcmp(entry->lname, sbuf) != 0) {
            HX_strclone(&entry->lname, sbuf);
        }
        HXbtree_add(lnlist, sbuf);
    }

    HXbtrav_free(travp);
    HXbtree_free(lnlist);
    printf("    %ld correction loops\n", stat_tries);
    printf("    %ld name(s) corrected\n", stat_users);
    return;
}

void sync_set_aflush(void) {
    if(Opt.c_aflush == AFLUSH_ON) {
        Q_SYSCTL(ACCDB_ADDFLAGS, ACCDB_AFLUSH);
    } else if(Opt.c_aflush == AFLUSH_DEFAULT) {
        unsigned long f = Q_SYSCTL(ACCDB_ADDFLAGS, 0);
        printf("Notice: AUTOFLUSH is turned %s\n",
         (f & ACCDB_AFLUSH) ? "on" : "off");
    } else {
        Q_SYSCTL(ACCDB_DELFLAGS, ACCDB_AFLUSH);
    }
    return;
}

FILE *sync_open_log(const char *file, char mc) {
    char mode[4] = {mc, '\0'};
    FILE *rv;

    if(mc == '\0') { return NULL; }
    if((rv = fopen(file, mode)) == NULL) {
        fprintf(stderr, "Error: Could not open %s: %s\n",
         file, strerror(errno));
        return NULL;
    }

    // change the number if the format should ever change
    fprintf(rv, "# $logformat 0\n" "# C-Spark %s\n", Opt.progv);
    return rv;
}

int sync_add(struct accdb_group *dgroup, struct HXbtree *al, FILE *logp) {
    // Adding the users to the system database
    struct HXbtree_node *node;
    long ic = al->itemcount;
    char *plain_pw;
    void *travp;
    int eax;

    if(ic == 0) { return 1; } // Nothing to do

    if(Opt.master_preadd != NULL) {
        struct HX_repmap catalog[] = {{'n', "%ld", &ic}, {0}};
        char cmd[MAXLNLEN];
        HX_strrep(cmd, MAXLNLEN, Opt.master_preadd, catalog);
        system(cmd);
    }

    plain_pw = malloc(Opt.i_pswdlen + 1);
    travp = HXbtrav_init(al, NULL);
    while((node = HXbtraverse(travp)) != NULL) {
        struct ds_entry *in = node->data;
        struct accdb_user out;
        char uhome[MAXFNLEN];

        memset(&out, 0, sizeof(struct accdb_user));
        out.lname = in->lname;
        out.uid   = -1;
        out.gid   = dgroup->gid;
        out.gecos = node->key;
        out.home  = make_home(in->lname, uhome, MAXFNLEN);
        out.shell = Opt.f_shell;
        out.igrp  = dgroup->gname;
        out.sgrp  = NULL;

        if(Opt.i_pswdlen < 0) {
            out.passwd = "!";
            out.last_change = 0;
        } else if(Opt.i_pswdlen == 0) {
            out.passwd = "";
            out.last_change = now_in_days();
        } else {
            unsigned long fl = GENPW_ONE_DIGIT | GENPW_ONE_CASE;
            if(Opt.b_pwphon) { fl |= GENPW_PHONEMIC; }
            accdb_genpw(plain_pw, Opt.i_pswdlen, fl);
            accdb_cryptpw(plain_pw, NULL, Opt.c_crymeth, &out.passwd);
            out.last_change = now_in_days();
        }

        out.keep_min    = 0;
        out.keep_max    = 366;
        out.warn_age    = 21;
        out.expire      = -1;
        out.inactive    = -1;

        if(Opt.user_preadd != NULL) {
            /* Order of positional parameters is important, as we should stay
            compatible in future. See vetc/spark.cfg or doc/spark_ui.html for
            the assigned order. */
            char cmd[MAXLNLEN];
            struct HX_repmap catalog[] = {
                {'l', NULL, out.lname}, {'n', NULL, out.gecos},
                {'u', "%lu", &out.uid}, {'g', "%lu", &out.gid},
                {'G', NULL, out.igrp},  {'S', NULL, out.sgrp},
                {'h', NULL, out.home},  {'s', NULL, out.shell},
                {0},
            };
            HX_strrep(cmd, MAXLNLEN, Opt.user_preadd, catalog);
            system(cmd);
        }

        if((eax = Q_USERADD(&out)) <= 0) {
            memset(plain_pw, '\0', Opt.i_pswdlen + 1);
            free(plain_pw);
            printf("\n");
            fprintf(stderr, "useradd: Something went wrong. (eax=%d,"
             " errno=%d: %s)\n", eax, errno, strerror(errno));
            HXbtrav_free(travp);
            return 0;
        }

        if(logp == NULL) {
            printf("Added %s with password %s\n", out.lname, plain_pw);
        } else {
            fprintf(logp, "+:%s:%s:%s:%s:%s\n", in->sgroup, in->nname,
             in->vname, out.lname, plain_pw);
        }

        if(Opt.i_pswdlen > 0) { free(out.passwd); }

        // Create home directory and optionally copy skeleton dir
        if(!create_home(out.home, out.uid, out.gid)) {
            printf("\n");
            fprintf(stderr, "Warning: Could not create home directory"
             "%s: %s\n", out.home, strerror(errno));
        }

        if(Opt.user_postadd != NULL) {
            char cmd[MAXLNLEN];
            struct HX_repmap catalog[] = {
                {'l', NULL, out.lname}, {'n', NULL, out.gecos},
                {'u', "%lu", &out.uid}, {'g', "%lu", &out.gid},
                {'G', NULL, out.igrp},  {'S', NULL, out.sgrp},
                {'h', NULL, out.home},  {'s', NULL, out.shell},
                {0},
            };
            HX_strrep(cmd, MAXLNLEN, Opt.user_postadd, catalog);
            system(cmd);
        }

        if((--ic & 0x3F) == 0) {
            printf("add: %ld user(s) remaining\n", ic);
        }
    }

    printf("Successfully added %lu user(s)\n", al->itemcount);
    memset(plain_pw, '\0', Opt.i_pswdlen + 1);
    free(plain_pw);
    HXbtrav_free(travp);

    if(Opt.master_postadd != NULL) {
        char cmd[MAXLNLEN];
        struct HX_repmap catalog[] = {{'n', "%lu", &al->itemcount}, {0}};
        HX_strrep(cmd, MAXLNLEN, Opt.master_postadd, catalog);
        system(cmd);
    }

    return 1;
}

int sync_del(struct HXdeque *dl, FILE *logp) {
    // Deleting the old users
    struct HXdeque_node *travp;
    long ic = dl->itemcount;
    char cdate[MAXSNLEN];
    int eax;

    if(ic == 0) { return 1; } // Nothing to do

    /* We generate a time stamp here so that a possible backup operation will
    write into the same directory for all users to be deleted. */
    now_in_ymdhms(cdate, MAXSNLEN);

    if(Opt.master_predel != NULL) {
        char cmd[MAXLNLEN];
        struct HX_repmap catalog[] =
         {{'d', NULL, cdate}, {'n', "%ld", &ic}, {0}};
        HX_strrep(cmd, MAXLNLEN, Opt.master_predel, catalog);
        system(cmd);
    }

    travp = dl->first;
    while(travp != NULL) {
        struct accdb_user rq;
        char *lname;

        memset(&rq, 0, sizeof(struct accdb_user));
        rq.lname = travp->ptr;
        rq.uid = rq.gid = -1;

        if((eax = Q_USERINFO(&rq, &rq, 1)) < 0) {
            printf("\n");
            fprintf(stderr, "userinfo: Something went wrong. (eax=%d, "
             "errno=%d: %s)\n", eax, errno, strerror(errno));
            return 0;
        } else if(eax == 0) {
            // This should not happen. Someone deleted the user before we did.
            printf("\n");
            fprintf(stderr, "Warning: Someone deleted \"%s\" before we"
             " did.\n", rq.lname);
            continue;
        }

        if(Opt.user_predel != NULL) {
            char cmd[MAXLNLEN];
            struct HX_repmap catalog[] = {
                {'l', NULL, rq.lname}, {'u', "%lu", &rq.uid},
                {'g', "%lu", &rq.gid}, {'G', NULL, rq.igrp},
                {'h', NULL, rq.home}, {'d', NULL, cdate}, {0},
            };
            HX_strrep(cmd, MAXLNLEN, Opt.user_predel, catalog);
            system(cmd);
        }
        HX_rrmdir(rq.home);

        // need a copy since travp->ptr will be undefined after userdel
        lname = HX_strdup(travp->ptr);

        // (surprise, we can reuse "rq" without modification)
        if((eax = Q_USERDEL(&rq)) < 0) {
            printf("\n");
            fprintf(stderr, "userdel: Something went wrong. (eax=%d, "
             "errno=%d: %s)\n", eax, errno, strerror(errno));
            return 0;
        }

        if(logp != NULL) { fprintf(logp, "-:%s\n", lname); }

        if(Opt.user_postdel != NULL) {
            char cmd[MAXLNLEN];
            struct HX_repmap catalog[] =
             {{'l', NULL, lname}, {'d', NULL, cdate}, {0}};
            HX_strrep(cmd, MAXLNLEN, Opt.user_postdel, catalog);
            system(cmd);
        }

        free(lname);
        if((--ic & 0x3F) == 0) {
            printf("delete: %ld user(s) remaining\n", ic);
        }
        travp = travp->Next;
    }

    printf("Successfully deleted %lu user(s)\n", dl->itemcount);

    if(Opt.master_postdel != NULL) {
        char cmd[MAXLNLEN];
        struct HX_repmap catalog[] =
         {{'d', NULL, cdate}, {'n', "%lu", &dl->itemcount}, {0}};
        HX_strrep(cmd, MAXLNLEN, Opt.master_postdel, catalog);
        system(cmd);
    }

    return 1;
}

void sync_free(struct HXbtree *al, struct HXdeque *dl) {
    if(al != NULL) {
        void *travp = HXbtrav_init(al, NULL);
        struct HXbtree_node *node;

        while((node = HXbtraverse(travp)) != NULL) {
            free_addl_entry(node->data);
        }

        HXbtrav_free(travp);
        HXbtree_free(al);
    }

    if(dl != NULL) { HXdeque_free(dl); }
    return;
}

//-----------------------------------------------------------------------------
inline static void free_addl_entry(struct ds_entry *e) {
    if(e == NULL) { return; }
    free(e->lname);
    free(e->vname);
    free(e->nname);
    free(e->rname);
    free(e->sgroup);
    free(e->xuid);
    free(e);
    return;
}

static void handle_user(struct accdb_user *p, struct HXbtree *al,
 struct HXdeque *dl, long *ukeep)
{
    char buf[MAXLNLEN], *brkp;

    buf[MAXLNLEN - 1] = '\0';
    strncpy(buf, p->gecos, MAXLNLEN - 1);

    if((brkp = strchr(buf, ',')) != NULL &&
     (brkp = strchr(++brkp, ',')) != NULL) {
        // strip further fields
        *brkp = '\0';
    }

    if(HXbtree_find(al, buf) != NULL) {
        ++*ukeep;
        free_addl_entry(HXbtree_get(al, buf));
        HXbtree_del(al, buf);
        // what remains in "al" is the users to add
    } else {
        // not anymore in data source
        HXdeque_push(dl, p->lname);
    }

    return;
}

//==[ End of file ]============================================================
