/*=============================================================================
Vitalnix User Management Suite
cui/xml_in.c - XML data source handling
  Copyright © Jan Engelhardt <jengelh [at] linux01 gwdg de>, 2003 - 2005
  -- License restrictions apply (LGPL v2)

  This file is part of Vitalnix.
  Vitalnix is free software; you can redistribute it and/or modify it
  under the terms of the GNU Lesser General Public License as published
  by the Free Software Foundation; however ONLY version 2 of the License.

  Vitalnix is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
  General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this program kit; if not, write to:
  Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
  02111-1307, USA.

  -- For details see the LICENSE file
=============================================================================*/
#include <libxml/parser.h>
#include <errno.h>
#include <iconv.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "cui/xml_in.h"
#define ICONV_NULL ((iconv_t)(-1))

struct state {
    xmlDoc *doc;
    xmlNode *ptr;
    iconv_t cd;
};

static char *convert(iconv_t, char *);

//-----------------------------------------------------------------------------
int xmlds_open(const char *fn, void **state) {
    struct state *sp;
    /*
        LOCAL ENCODING
        This will probably be "iso-8859-1" for most Windows installations
        and some Linux distributions that do not yet support UTF-8.
        Use NULL to use UTF-8.
    */
    const char *local_encoding = NULL;
    if((sp = *state = malloc(sizeof(struct state))) == NULL) {
        return -errno;
    }

    xmlKeepBlanksDefault(0);

    /* Need iconv() since libxml stores everything in UTF-8 internally,
    but if local encoding is not yet there.... */
    if(local_encoding != NULL) {
        sp->cd = iconv_open(local_encoding, "utf-8");
        if(sp->cd == ICONV_NULL) {
            free(sp);
            return -errno;
        }
    } else {
        sp->cd = ICONV_NULL;
    }

    /* xmlParseFile() already scans the file completely and stores the
    structure tree plus components in memory. */
    if((sp->doc = xmlParseFile(fn)) == NULL) {
        xmlds_close(*state);
        return 0;
    }

    if((sp->ptr = xmlDocGetRootElement(sp->doc)) == NULL ||
     strcmp(sp->ptr->name, "datasource") != 0) {
        xmlds_close(*state);
        return 0;
    }

    sp->ptr = sp->ptr->children;
    return 1;
}

int xmlds_read(void *state, struct ds_entry *x) {
    /* The reading process is very simple, since libxml takes all the pain to
    parse the file down, and its API directly provides us with the stuff we
    want. */
    struct state *sp = state;
    if(sp->ptr == NULL) { return 0; }
    if(strcmp(sp->ptr->name, "user") == 0) {
        x->vname = convert(sp->cd, xmlGetProp(sp->ptr, "vname"));

        if(x->vname == NULL) { // bogus entry
            sp->ptr = sp->ptr->next;
            return -EINVAL;
        }
        if(*x->vname == '\0') { // bogus entry
            free(x->vname);
            sp->ptr = sp->ptr->next;
            return -EINVAL;
        }

        x->nname  = convert(sp->cd, xmlGetProp(sp->ptr, "nname"));
        x->sgroup = convert(sp->cd, xmlGetProp(sp->ptr, "sgroup"));
        x->xuid   = convert(sp->cd, xmlGetProp(sp->ptr, "xuid"));
    }

    sp->ptr = sp->ptr->next;
    return 1;
}

void xmlds_close(void *state) {
    struct state *sp = state;
    if(sp->doc != NULL) { xmlFreeDoc(sp->doc); }
    xmlCleanupParser();
    if(sp->cd != ICONV_NULL) { iconv_close(sp->cd); }
    free(sp);
    return;
}

//-----------------------------------------------------------------------------
static char *convert(iconv_t cd, char *in) {
    char *inp = in, *out, *outp;
    size_t inl, outl;

    if(cd == ICONV_NULL) {
        // We want UTF-8, and "in" already is
        return in;
    }
    if(in == NULL) {
        return NULL;
    }
    if(*in == '\0') {
        xmlFree(in);
        return NULL;
    }

    if((outp = out = malloc(inl = outl = 2 * strlen(in) + 1)) == NULL) {
        return NULL;
    }
    iconv(cd, &inp, &inl, &outp, &outl);
    xmlFree(in);
    return out;
}

//==[ End of file ]============================================================
