/*=============================================================================
Vitalnix User Management Suite
supply/accdbinfo.c - Info about ACCDB back-end modules
  Copyright © Jan Engelhardt <jengelh [at] linux01 gwdg de>, 2003 - 2005
  -- License restrictions apply (LGPL v2)

  This file is part of Vitalnix.
  Vitalnix is free software; you can redistribute it and/or modify it
  under the terms of the GNU Lesser General Public License as published
  by the Free Software Foundation; however ONLY version 2 of the License.

  Vitalnix is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
  General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this program kit; if not, write to:
  Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
  02111-1307, USA.

  -- For details see the LICENSE file
=============================================================================*/
#include <popt.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <libHX.h>
#include <accdb.h>
#include "sysprog/shared.h"

static void backend_info(char *);
static void read_ldso_conf(void);
static void read_environment(void);
static int get_options(int *, const char ***);

static struct HXdeque *Dirs = NULL;
static size_t mcount = 0;
static int OP_open = 0;

//-----------------------------------------------------------------------------
int main(int argc, char **argv) {
    struct HXdeque_node *cd;

    Dirs = HXdeque_init(0);
    HXdeque_push(Dirs, "/lib");
    HXdeque_push(Dirs, "/usr/lib");
    read_ldso_conf();
    HXdeque_push(Dirs, ".");
    read_environment();

    get_options(&argc, (const char ***)&argv);
    backend_info("*");

    cd = Dirs->first;
    while(cd != NULL) {
        struct HXdir *cdp;
        char *dentry;

        if((cdp = HXdir_open(cd->ptr)) == NULL) {
          cd = cd->Next;
          continue;
        }

        while((dentry = HXdir_read(cdp)) != NULL) {
            char buf[MAXFNLEN];
            if(strncmp(dentry, "accdb_", 6) != 0) { continue; }
            snprintf(buf, MAXLNLEN, "%s/%s", (char *)cd->ptr, dentry);
            backend_info(buf);
        }

        HXdir_close(cdp);
        cd = cd->Next;
    }

    printf("------------------------------------------------------------\n");
    printf("Read %u modules\n", (unsigned int)mcount);
    HXdeque_free(Dirs);
    return !1;
}

static void backend_info(char *fn) {
    struct accdb_module *md = accdb_load(fn);
    char sepl = '<', sepr = '>';

    if(md == NULL) { return; }
    if(strcmp(fn, "*") == 0) { sepl = sepr = '"'; }

    printf(
      "------------------------------------------------------------\n"
      "Module %c%s%c:\n"
      "    Name       : %s\n"
      "    Description: %s\n",
      sepl, fn, sepr, md->name, md->desc
    );

    if(strchr(md->info, '\n') == NULL) {
        printf("    Info       : %s\n", md->info);
    } else {
        printf("    Info:\n%s", md->info);
    }

    if(OP_open) {
        md->open(md);
        md->close(md);
    }

    accdb_unload(md);
    ++mcount;
    return;
}

static void read_ldso_conf(void) {
    FILE *fp;
    char buf[MAXLNLEN];

    if((fp = fopen("/etc/ld.so.conf", "r")) == NULL) { return; }
    while(fgets(buf, MAXLNLEN, fp) != NULL) {
        char *q;
        HX_chomp(buf);
        if((q = strchr(buf, '=')) != NULL) { *q = '\0'; }
        HXdeque_push(Dirs, HX_strdup(buf));
    }

    fclose(fp);
    return;
}

static void read_environment(void) {
    char *entry, *our, *travp;
    if((entry = getenv("LD_LIBRARY_PATH")) == NULL) { return; }
    HX_strclone(&our, entry);
    travp = our;

    while((travp = HX_strsep(&travp, ":")) != NULL) {
        HXdeque_push(Dirs, travp);
    }

    free(our);
    return;
}

static int get_options(int *argc, const char ***argv) {
    static const char *_empty_argv[] = {NULL};
    char *flag_l = NULL;

    const struct poptOption options_table[] = {
        {NULL, 'L', POPT_ARG_STRING, &flag_l, 'L',
         "Additional search directory", "dir"},
        {NULL, 'O', POPT_ARG_NONE, &OP_open, '\0',
         "Open (q_open) the ACCDB back-end", NULL},

        POPT_AUTOHELP
        POPT_TABLEEND
    };

    poptContext ctx;
    const char **args;
    int c, argk;

    ctx = poptGetContext(**argv, *argc, *argv, options_table, 0);
    while((c = poptGetNextOpt(ctx)) >= 0) {
        switch(c) {
            case 'L':
                HXdeque_push(Dirs, flag_l);
                break;
        }
    }

    if(c < -1) {
        fprintf(stderr, "%s: %s\n", poptBadOption(ctx, 0), poptStrerror(c));
        return 0;
    }

    if((args = poptGetArgs(ctx)) != NULL) {
        argk = SH_count_args(args);
        poptDupArgv(argk, args, argc, argv);
    } else {
        *argc = 0;
        *argv = _empty_argv;
    }

    poptFreeContext(ctx);
    return 1;
}

//==[ End of file ]============================================================
