/*=============================================================================
Vitalnix User Management Suite
sysprog/groupdel.c - Delete a group
  Copyright © Jan Engelhardt <jengelh [at] linux01 gwdg de>, 2003 - 2005
  -- License restrictions apply (LGPL v2)

  This file is part of Vitalnix.
  Vitalnix is free software; you can redistribute it and/or modify it
  under the terms of the GNU Lesser General Public License as published
  by the Free Software Foundation; however ONLY version 2 of the License.

  Vitalnix is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
  General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this program kit; if not, write to:
  Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
  02111-1307, USA.

  -- For details see the LICENSE file
=============================================================================*/
#include <errno.h>
#include <popt.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <libHX.h>
#include <accdb.h>
#include "sysprog/shared.h"

enum {
    E_SUCCESS = 0,
    E_OTHER,      // other error, see errno
    E_OPEN,       // unable to open back-end module or DB
    E_NOEXIST,    // group does not exist
    E_UPDATE,     // db->groupdel() did not return ok
    E_CLOSE,      // db->close() did not return ok
    E_PRIMARY,    // will not delete a user's primary group, unless -F is given
};

struct {
    char *gname;
    // groupdel internal
    char *ac_after, *ac_before;
    int force;
} Opt = {
    gname: NULL,
    ac_after: NULL, ac_before: NULL, force: 0,
};

static int check_pri_group(struct accdb_module *, struct accdb_group *);
static int get_options(int *, const char ***);
static void read_config(void);

static char *module_path = "*";

//-----------------------------------------------------------------------------
int main(int argc, char **argv) {
    struct accdb_module *db;
    struct accdb_group mmask;
    int eax, rv = E_SUCCESS;
    struct HX_repmap sr_map[] = {{'G', NULL, mmask.gname}, {0}};

    memset(&mmask, 0, sizeof(struct accdb_group));
    read_config();
    if(!(eax = get_options(&argc, (const char ***)&argv))) { return E_OTHER; }

    // ----------------------------------------
    if((db = accdb_load(module_path)) == NULL) {
        fprintf(stderr, "Could not load ACCDB back-end module \"%s\": %s\n",
         module_path, strerror(errno));
        return E_OPEN;
    }

    if((eax = db->open(db)) <= 0) {
        fprintf(stderr, "Could not open ACCDB back-end: %s"
         " (eax=%d, errno=%d)\n", strerror(errno), eax, errno);
        rv = E_OPEN;
        goto __main__close_accdb;
    }

    memset(&mmask, 0, sizeof(struct accdb_group));
    mmask.gid = -1;
    mmask.gname = Opt.gname = *argv;

    // ----------------------------------------
    if((eax = db->groupinfo(db, &mmask, NULL, 0)) < 0) {
        fprintf(stderr, "Error querying the ACCDB: %s (eax=%d, errno=%d)\n",
         strerror(errno), eax, errno);
        rv = E_OTHER;
        goto __main__close_backend;
    } else if(eax == 0) {
        fprintf(stderr, "Group \"%s\" does not exist\n", mmask.gname);
        rv = E_NOEXIST;
        goto __main__close_backend;
    }

    if(!Opt.force && check_pri_group(db, &mmask)) {
        fprintf(stderr, "Will not remove a user's primary group\n");
        rv = E_PRIMARY;
        goto __main__close_backend;
    }

    // ----------------------------------------
    if(Opt.ac_before != NULL) { SH_runcmd(Opt.ac_before, sr_map); }

    if((eax = db->groupdel(db, &mmask)) <= 0) {
        fprintf(stderr, "Error: Deleting group failed: %s"
         " (eax=%d, errno=%d)\n", strerror(errno), eax, errno);
        rv = E_UPDATE;
    } else if(Opt.ac_after != NULL) {
        SH_runcmd(Opt.ac_after, sr_map);
    }

    // ----------------------------------------
__main__close_backend:
    if((eax = db->close(db)) <= 0) {
        fprintf(stderr, "Warning: DB was not cleanly closed:"
         " %s (eax=%d, errno=%d)\n", strerror(errno), eax, errno);
        if(rv == E_SUCCESS) { rv = E_CLOSE; }
    }

__main__close_accdb:
    accdb_unload(db);
    return rv;
}

//-----------------------------------------------------------------------------
static int check_pri_group(struct accdb_module *db, struct accdb_group *mm) {
    struct accdb_user u;
    int pg = 0; // some user has this as primary group

    while(db->usertrav(db, &u) > 0) {
        if((u.gid != -1 && u.gid == mm->gid) || (u.igrp != NULL &&
         mm->gname != NULL && strcmp(u.igrp, mm->gname) == 0)) {
            pg = 1;
            break;
        }
    }

    return pg;
}

static int get_options(int *argc, const char ***argv) {
    static const char *_empty_argv[] = {NULL};
    const struct poptOption options_table[] = {
        // New, Vitalnix-userdel options
        {NULL, 'A', POPT_ARG_STRING | POPT_ARGFLAG_OPTIONAL, &Opt.ac_after,
         '\0', "Program to run after group deletion", "cmd"},
        {NULL, 'B', POPT_ARG_STRING | POPT_ARGFLAG_OPTIONAL, &Opt.ac_before,
         '\0', "Program to run before group deletion", "cmd"},
        {NULL, 'F', POPT_ARG_NONE, &Opt.force, '\0',
         "Force deletion of group even if users have it as"
         " primary group", NULL},
        {NULL, 'M', POPT_ARG_STRING, &module_path, '\0',
         "Use a different module than \"*\" (the default)", "name"},
        POPT_AUTOHELP
        POPT_TABLEEND
    };

    poptContext ctx;
    const char **args;
    int c, argk;

    ctx = poptGetContext(**argv, *argc, *argv, options_table, 0);
    poptSetOtherOptionHelp(ctx, "[OPTIONS...] group");
    while((c = poptGetNextOpt(ctx)) >= 0);
    if(c < -1) {
        fprintf(stderr, "%s: %s: %s\n", **argv, poptBadOption(ctx, 0),
         poptStrerror(c));
        poptFreeContext(ctx);
        return 0;
    }

    if((args = poptGetArgs(ctx)) != NULL) {
        argk = SH_count_args(args);
        poptDupArgv(argk, args, argc, argv);
    } else {
        *argc = 0;
        *argv = _empty_argv;
    }

    if(args == NULL || *args == NULL) {
        // gname is mandatory
        poptPrintUsage(ctx, stderr, 0);
        poptFreeContext(ctx);
        return 0;
    }

    poptFreeContext(ctx);
    return 1;
}

static void read_config(void) {
    static struct rconfig_opt options_table[] = {
        {"AC_BEFORE", RCONF_STRING, &Opt.ac_before, NULL},
        {"AC_AFTER",  RCONF_STRING, &Opt.ac_after,  NULL},
        {NULL},
    };

    HX_rconfig_pv(Conf_paths, "groupdel.conf", options_table, RCONFIG_ONE);
    return;
}

//==[ End of file ]============================================================
