/*=============================================================================
Vitalnix User Management Suite
sysprog/userdel.c - Delete a user account and related files
  Copyright © Jan Engelhardt <jengelh [at] linux01 gwdg de>, 2003 - 2005
  -- License restrictions apply (LGPL v2)

  This file is part of Vitalnix.
  Vitalnix is free software; you can redistribute it and/or modify it
  under the terms of the GNU Lesser General Public License as published
  by the Free Software Foundation; however ONLY version 2 of the License.

  Vitalnix is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
  General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this program kit; if not, write to:
  Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
  02111-1307, USA.

  -- For details see the LICENSE file
=============================================================================*/
#include <errno.h>
#include <popt.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <libHX.h>
#include <accdb.h>
#include "sysprog/shared.h"

enum {
    E_SUCCESS = 0,
    E_OTHER,       // other error, see errno
    E_OPEN,        // unable to open back-end module or DB
    E_NOEXIST,     // user does not exist
    E_DENY,        // will not remove 'root' or UID 0 without -F (force)
    E_UPDATE,      // db->userdel() did not return ok
    E_CLOSE,       // db->close() did not return ok
};

struct {
    char *lname;
    // userdel internal
    char *ac_after, *ac_before;
    int force, rm_home;
} Opt = {
    lname: NULL, ac_after: NULL, ac_before: NULL, rm_home: 0,
};

static int count_oc(const char *, char);
static int get_options(int *, const char ***);
static void read_config(void);

static char *module_path = "*";

//-----------------------------------------------------------------------------
int main(int argc, char **argv) {
    struct accdb_module *db;
    struct accdb_user mmask, result;
    int eax, rv = E_SUCCESS;
    char *homedir = NULL;
    struct HX_repmap sr_map[] = {{'l', NULL, mmask.lname}, {0}};

    memset(&mmask, 0, sizeof(struct accdb_user));
    read_config();
    if(!(eax = get_options(&argc, (const char ***)&argv))) { return E_OTHER; }

    // ----------------------------------------
    if((db = accdb_load(module_path)) == NULL) {
        fprintf(stderr, "Could not load ACCDB back-end module \"%s\": %s\n",
         module_path, strerror(errno));
        return E_OPEN;
    }

    if((eax = db->open(db)) <= 0) {
        fprintf(stderr, "Could not open ACCDB back-end: %s"
         " (eax=%d, errno=%d)\n", strerror(errno), eax, errno);
        rv = E_OPEN;
        goto __main__close_accdb;
    }

    memset(&mmask, 0, sizeof(struct accdb_user));
    mmask.uid = mmask.gid = -1;
    mmask.lname = Opt.lname = *argv;

    // ----------------------------------------
    if((eax = db->userinfo(db, &mmask, &result, 1)) < 0) {
        fprintf(stderr, "Error querying the ACCDB: %s (eax=%d, errno=%d)\n",
         strerror(errno), eax, errno);
        rv = E_OTHER;
        goto __main__close_backend;
    } else if(eax == 0) {
        fprintf(stderr, "User \"%s\" does not exist\n", mmask.lname);
        rv = E_NOEXIST;
        goto __main__close_backend;
    }

    if(!Opt.force && (strcmp(result.lname, "root") == 0 || result.uid == 0)) {
        fprintf(stderr, "Will not remove user \"root\" or user with UID 0,"
         " unless -F is given.");
        rv = E_DENY;
        goto __main__close_backend;
    }

    // ----------------------------------------
    if(Opt.ac_before != NULL) { SH_runcmd(Opt.ac_before, sr_map); }

    if(Opt.rm_home) {
        char buf[MAXFNLEN];
        snprintf(buf, MAXFNLEN, "/var/spool/mail/%s", mmask.lname);
        unlink(buf);
        homedir = HX_strdup(result.home);
    }

    if((eax = db->userdel(db, &mmask)) <= 0) {
        fprintf(stderr, "Error: Deleting user failed: %s (eax=%d, errno=%d)\n",
         strerror(errno), eax, errno);
        rv = E_UPDATE;
    } else if(Opt.ac_after != NULL) {
        SH_runcmd(Opt.ac_after, sr_map);
    }

    if(Opt.rm_home) {
        if(homedir == NULL || *homedir == '\0') {
            fprintf(stderr, "Warning: User had no home directory. Not removing"
             " anything.\n");
        } else if(strcmp(homedir, "/") == 0) {
            fprintf(stderr, "Warning: Will refuse to delete home directory"
             " \"/\"\n");
        } else if(count_oc(homedir, '/') < 2 && !Opt.force) {
            fprintf(stderr, "Warning: Will not remove home directory \"%s\""
             " which has only two slashes and thus looks like a system"
             " directory. (Remove it manually.)\n", homedir);
        } else {
            HX_rrmdir(homedir);
        }

        free(homedir);
    }

    // ----------------------------------------
__main__close_backend:
    if((eax = db->close(db)) <= 0) {
        fprintf(stderr, "Warning: DB was not cleanly closed:"
         " %s (eax=%d, errno=%d)\n", strerror(errno), eax, errno);
        if(rv == E_SUCCESS) { rv = E_CLOSE; }
    }

__main__close_accdb:
    accdb_unload(db);
    return rv;
}

//-----------------------------------------------------------------------------
static int count_oc(const char *fn, char s) {
    const char *ptr = fn;
    int n = 0;
    if(fn == NULL) { return 0; }

    while((ptr = strchr(ptr, '/')) != NULL) {
        ++n;
        ++ptr;
    }

    return n;
}

static int get_options(int *argc, const char ***argv) {
    static const char *_empty_argv[] = {NULL};
    const struct poptOption options_table[] = {
        // New, accdb-userdel options
        {NULL, 'F', POPT_ARG_NONE, &Opt.force, '\0',
         "Force deletion even if UID is 0 or name is 'root'", NULL},
        {NULL, 'M', POPT_ARG_STRING, &module_path, '\0',
         "Use a different module than \"*\" (the default)", "name"},

        // Default options
        {NULL, 'r', POPT_ARG_NONE, &Opt.rm_home, 'r',
         "Remove: home directory and mail spool", NULL},

        POPT_AUTOHELP
        POPT_TABLEEND
    };

    poptContext ctx;
    const char **args;
    int c, argk;

    ctx = poptGetContext(**argv, *argc, *argv, options_table, 0);
    while((c = poptGetNextOpt(ctx)) >= 0);

    if(c < -1) {
        fprintf(stderr, "%s: %s\n", poptBadOption(ctx, 0), poptStrerror(c));
        poptFreeContext(ctx);
        return 0;
    }

    if((args = poptGetArgs(ctx)) != NULL) {
        argk = SH_count_args(args);
        poptDupArgv(argk, args, argc, argv);
    } else {
        *argc = 0;
        *argv = _empty_argv;
    }

    if(args == NULL || *args == NULL) {
        // loginname is mandatory
        poptPrintUsage(ctx, stderr, 0);
        poptFreeContext(ctx);
        return 0;
    }

    poptFreeContext(ctx);
    return 1;
}

static void read_config(void) {
    static struct rconfig_opt options_table[] = {
        {"AC_BEFORE", RCONF_STRING, &Opt.ac_before, NULL},
        {"AC_AFTER",  RCONF_STRING, &Opt.ac_after,  NULL},
        {NULL},
    };

    HX_rconfig_pv(Conf_paths, "userdel.conf", options_table, RCONFIG_ONE);
    return;
}

//==[ End of file ]============================================================
